-- *************************************************************************
-- DISCLAIMER. THIS SOFTWARE WAS WRITTEN BY EMPLOYEES OF THE U.S.
-- GOVERNMENT AS A PART OF THEIR OFFICIAL DUTIES AND, THEREFORE, IS NOT
-- PROTECTED BY COPYRIGHT. HOWEVER, THIS SOFTWARE CODIFIES THE FINALIST
-- CANDIDATE ALGORITHMS (i.e., MARS, RC6tm, RIJNDAEL, SERPENT, AND
-- TWOFISH) IN THE ADVANCED ENCRYPTION STANDARD (AES) DEVELOPMENT EFFORT
-- SPONSORED BY THE NATIONAL INSTITUTE OF STANDARDS AND TECHNOLOGY (NIST)
-- AND MAY BE PROTECTED BY ONE OR MORE FORMS OF INTELLECTUAL PROPERTY. THE
-- U.S. GOVERNMENT MAKES NO WARRANTY, EITHER EXPRESSED OR IMPLIED,
-- INCLUDING BUT NO LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY
-- OR FITNESS FOR A PARTICULAR PURPOSE, REGARDING THIS SOFTWARE. THE U.S.
-- GOVERNMENT FURTHER MAKES NO WARRANTY THAT THIS SOFTWARE WILL NOT
-- INFRINGE ANY OTHER UNITED STATES OR FOREIGN PATENT OR OTHER
-- INTELLECTUAL PROPERTY RIGHT. IN NO EVENT SHALL THE U.S. GOVERNMENT BE
-- LIABLE TO ANYONE FOR COMPENSATORY, PUNITIVE, EXEMPLARY, SPECIAL,
-- COLLATERAL, INCIDENTAL, CONSEQUENTIAL, OR ANY OTHER TYPE OF DAMAGES IN
-- CONNECTION WITH OR ARISING OUT OF COPY OR USE OF THIS SOFTWARE.
-- *************************************************************************

-- ===========================================================================
-- File Name : alg_pipe.vhdl
-- Author    : NSA
-- Date      : December 1999
-- Project   : AES Candidate Evaluation - TWOFISH
-- Purpose   : This block implements the TWOFISH algorithm for a pipelined case
-- Notes     :
-- ===========================================================================

library ieee;
use ieee.std_logic_1164.all;
use WORK.twofish_pack.all;

-- ===========================================================================
-- =========================== Interface Description =========================
-- ===========================================================================

entity ALG_PIPE is

  port (clock        :  in STD_LOGIC;        -- clock signal
        reset        :  in STD_LOGIC;        -- active high reset (asynchronous)

        ALG_DATAIN   :  in SLV_128;          -- input data
        ALG_KEY_ENC  :  in KEY_ARRAY_TYPE;   -- array of encrypt sub-keys
        ALG_KEY_DEC  :  in KEY_ARRAY_TYPE;   -- array of decrypt sub-keys
        ALG_CVSIZE   :  in CVSIZE_PIPE_TYPE; -- pipelined CV size
        ALG_S        :  in S_PIPE_TYPE;      -- pipelined S values
        ALG_ENC      :  in std_logic;        -- (1 = encrypt, 0 = decrypt)

        ALG_DATAOUT  :  out SLV_128          -- Processed output data
  );

end ALG_PIPE;


architecture ALG_PIPE_RTL of ALG_PIPE is


-- ===========================================================================
-- ========================== Component Definition ===========================
-- ===========================================================================

component REG128B 

  port (clock    :  in std_logic;   -- clock signal
        reset    :  in std_logic;   -- active high reset (asynchronous)

        DATA_IN  :  in SLV_128;     -- input data bus
        DATA_OUT :  out SLV_128     -- output data bus

  );

end component;


component TWOFISH_ROUND

  port (datain       :  in SLV_128;
        enc_key1     :  in SLV_32;
        enc_key2     :  in SLV_32;
        dec_key1     :  in SLV_32;
        dec_key2     :  in SLV_32;
        cv_length    :  in CV_SIZE_TYPE;
        S            :  in S_ARRAY_TYPE;
        round        :  in SLV_5;
        enc_dec_b    :  in std_logic;
        dataout      :  out SLV_128
  );

end component;


type ROUND_VEC_TYPE is array (0 to 15) of SLV_5;
constant ROUND_VEC : ROUND_VEC_TYPE := (
  "00001", "00010", "00011", "00100",
  "00101", "00110", "00111", "01000",
  "01001", "01010", "01011", "01100",
  "01101", "01110", "01111", "10000" );




-- ===========================================================================
-- =========================== Signal Definition =============================
-- ===========================================================================

signal ALG_DATA        : PIPE_DATA_TYPE;  -- register output
signal ALG_DATA_INT    : PIPE_DATA_TYPE;  -- procedure call to reg interconnect
signal ROUND_STEP      : ROUND_TYPE;      -- indicates current algorithm step
signal ROUND_NUM       : ROUND_VEC_TYPE;

begin

-- ===========================================================================
-- ======================== Generate Pipe Structure ==========================
-- ===========================================================================
--
-- PURPOSE:  
--
-- The following generate statements create a pipelined architecture for 
-- the encryption/decryption.


  ROUND_NUM <= ROUND_VEC;

G1: for ROUND_STEP in FIRST_ROUND to LAST_ROUND generate

-- =========================== First Round ===================================
-- Generate first round connections

   G1a:  if (ROUND_STEP = FIRST_ROUND) generate

      INIT : PRE_ADD( ALG_DATAIN,
                         ALG_KEY_ENC(0),
                         ALG_KEY_ENC(1),
                         ALG_KEY_ENC(2),
                         ALG_KEY_ENC(3),
                         ALG_KEY_DEC(0),
                         ALG_KEY_DEC(1),
                         ALG_KEY_DEC(2),
                         ALG_KEY_DEC(3),
                         ALG_ENC,
                         ALG_DATA_INT(ROUND_STEP) );      -- data interconnect

      DATAREG : REG128B port map ( clock,
                                   reset,
                                   ALG_DATA_INT(ROUND_STEP),   -- interconnect
                                   ALG_DATA(ROUND_STEP) );     -- output data

   end generate;  -- G1a

-- ========================== Round i ========================================
-- Generate middle rounds for creating encryption/decryption.

   G1b:  if (ROUND_STEP < LAST_ROUND and ROUND_STEP > 0) generate


      STEP : TWOFISH_ROUND port map ( ALG_DATA(ROUND_STEP - 1) ,  -- input data
            	        ALG_KEY_ENC(2 * ROUND_STEP + 6), 
             	        ALG_KEY_ENC(2 * ROUND_STEP + 7),
              	        ALG_KEY_DEC(2 * ROUND_STEP + 6),
               	        ALG_KEY_DEC(2 * ROUND_STEP + 7),
                        ALG_CVSIZE(ROUND_STEP),
                        ALG_S(ROUND_STEP),
                        ROUND_NUM(ROUND_STEP-1),
                        ALG_ENC,
                        ALG_DATA_INT(ROUND_STEP) );            -- output data

      DATAREG : REG128B port map ( clock,
                                   reset,
                                   ALG_DATA_INT(ROUND_STEP),   -- interconnect
                                   ALG_DATA(ROUND_STEP) );     -- output data

   end generate;  -- G1b

-- ========================= Last Round ======================================
-- Generate last round and special case connections

   G1c:  if (ROUND_STEP = LAST_ROUND) generate

      POST : POST_ADD( ALG_DATA(ROUND_STEP - 1),
                          ALG_KEY_ENC(4),
                          ALG_KEY_ENC(5),
                          ALG_KEY_ENC(6),
                          ALG_KEY_ENC(7),
                          ALG_KEY_DEC(4),
                          ALG_KEY_DEC(5),
                          ALG_KEY_DEC(6),
                          ALG_KEY_DEC(7),
                          ALG_ENC,
                          ALG_DATA_INT(ROUND_STEP) );


      DATAREG : REG128B port map ( clock,
                                   reset,
                                   ALG_DATA_INT(ROUND_STEP),
                                   ALG_DATA(ROUND_STEP) ); -- output data

   end generate; -- G1c

end generate;  -- G1

ALG_DATAOUT<= ALG_DATA(17);


end ALG_PIPE_RTL;

-- ===========================================================================
-- ============================ Configuration ================================
-- ===========================================================================

configuration CFG_ALG_PIPE of ALG_PIPE is

   for ALG_PIPE_RTL

   end for;

end CFG_ALG_PIPE;
